"""
    Module for running Docker images with Singularity.   
    Derived from https://github.com/BD2KGenomics/toil/blob/master/src/toil/lib/docker.py

    Contains two user-facing functions: singularityCall and singularityCheckOutput
    Example of using singularityCall in a Toil pipeline to index a FASTA file with SAMtools:
        def toil_job(job):
            work_dir = job.fileStore.getLocalTempDir()
            path = job.fileStore.readGlobalFile(ref_id, os.path.join(work_dir, 'ref.fasta')
            parameters = ['faidx', path]
            singularityCall(job, tool='quay.io/ucgc_cgl/samtools:latest', work_dir=work_dir, parameters=parameters)
"""
import base64
import hashlib
import logging
import subprocess
import pipes
import os
import pathlib
import shutil
import sys
import tempfile
import time

logger = logging.getLogger(__name__)

def is_containerized():
    """
    Return True if we think we are already running in a Docker/Kubernetes
    container (where Singularity is unlikely to work without user-mode
    namespaces), and False otherwsie.
    """

    if not os.path.exists('/proc/self/cgroup'):
        # Not on container-having Linux
        return False

    with open('/proc/self/cgroup') as fh:
        for line in fh:
            line = line.lower()
            if 'docker' in line or 'kube' in line:
                # If any of the cgroups smells Docker or Kube-y, assume we are
                # in a container.
                return True
    return False


def singularityCall(job,
               tool,
               parameters=None,
               workDir=None,
               singularityParameters=None,
               outfile=None,
               mount_list=None):
    """
    Throws CalledProcessorError if the Singularity invocation returns a non-zero exit code
    This function blocks until the subprocess call to Singularity returns
    :param toil.Job.job job: The Job instance for the calling function.
    :param str tool: Name of the Docker image to be used (e.g. quay.io/ucsc_cgl/samtools:latest).
    :param list[str] parameters: Command line arguments to be passed to the tool.
           If list of lists: list[list[str]], then treat as successive commands chained with pipe.
    :param str workDir: Directory to mount into the container via `-B`.
           Destination convention is /mnt, which almost certainly exists in the
           container.
    :param list[str] singularityParameters: Parameters to pass to Singularity.
           Overrides defaults which mount the workDir and configure user mode and
           writability.
    :param file outfile: Pipe output of Singularity call to file handle
    :param list[str] mount_list: List of directories from which to mount into the
           container via `-B`. Destination convention is /mnt, which almost certainly
           exists in the container.
    """
    return _singularity(job, tool=tool, parameters=parameters, workDir=workDir, singularityParameters=singularityParameters,
                        outfile=outfile, checkOutput=False, mount_list=mount_list)


def singularityCheckOutput(job,
                      tool,
                      parameters=None,
                      workDir=None,
                      singularityParameters=None,
                      mount_list=None):
    """
    Returns the stdout from the Singularity invocation (via subprocess.check_output)
    Throws CalledProcessorError if the Singularity invocation returns a non-zero exit code
    This function blocks until the subprocess call to Singularity returns
    :param toil.Job.job job: The Job instance for the calling function.
    :param str tool: Name of the Docker image to be used (e.g. quay.io/ucsc_cgl/samtools:latest).
    :param list[str] parameters: Command line arguments to be passed to the tool.
           If list of lists: list[list[str]], then treat as successive commands chained with pipe.
    :param str workDir: Directory to mount into the container via `-B`.
           Destination convention is /mnt, which almost certainly exists in the
           container.
    :param list[str] singularityParameters: Parameters to pass to Singularity.
           Overrides defaults which mount the workDir and configure user mode and
           writability.
    :param list[str] mount_list: List of directories from which to mount into the
           container via `-B`. Destination convention is /mnt, which almost certainly
           exists in the container.
    :returns: Stdout from the singularity call
    :rtype: str
    """
    return _singularity(job, tool=tool, parameters=parameters, workDir=workDir,
                   singularityParameters=singularityParameters, checkOutput=True, mount_list=mount_list)


def _singularity(job,
            tool,
            parameters=None,
            workDir=None,
            singularityParameters=None,
            outfile=None,
            checkOutput=False,
            mount_list=None):
    """
    :param toil.Job.job job: The Job instance for the calling function.
    :param str tool: Name of the Docker image to be used (e.g. quay.io/ucsc_cgl/samtools).
    :param list[str] parameters: Command line arguments to be passed to the tool.
           If list of lists: list[list[str]], then treat as successive commands chained with pipe.
    :param str workDir: Directory to mount into the container via `-B`.
           Destination convention is /mnt, which almost certainly exists in the
           container.
    :param list[str] singularityParameters: Parameters to pass to Singularity.
           Overrides defaults which mount the workDir and configure user mode and
           writability.
    :param file outfile: Pipe output of Singularity call to file handle
    :param bool checkOutput: When True, this function returns singularity's output.
    :param list[str] mount_list: List of directories from which to mount into the
           container via `-B`. Destination convention is /mnt, which almost certainly
           exists in the container.
    """
    if parameters is None:
        parameters = []
    if workDir is None:
        workDir = os.getcwd()

    # Setup the outgoing subprocess call for singularity.
    baseSingularityCall = ['singularity', 'exec']
    if singularityParameters:
        baseSingularityCall += singularityParameters
    else:
        # Make the container writable. Writing to the container is still not
        # advised, but some tools/environments break if it is not enabled.
        baseSingularityCall.append('-w')

        if is_containerized():
            # We are already in a container. We need to run in user mode,
            # because the container may not be privileged. If we don't use user
            # mode, Singularity tries to do some confining that it can't do in
            # an un-privileged container, and fails.
            baseSingularityCall.append('-u')
            
        if not str(pathlib.Path.home()).startswith('/home'):
            # Newer versions of Singularity will fail if they can't mount the
            # home directory, which they can't do if the directory the home
            # directory is in doesn't exist in the container. If it isn't just
            # under /home, assume it might not be in the container and tell
            # Singularity not to try to mount it. See
            # https://github.com/hpcng/singularity/issues/4995
            baseSingularityCall.append('--no-home')

        # Mount workdir as /mnt and work in there.
        # Hope the image actually has a /mnt available.
        # Otherwise this silently doesn't mount.
        # But with -u (user namespaces) we have no luck pointing in-container
        # home at anything other than our real home (like something under /var
        # where Toil puts things).
        # Note that we target Singularity 3+.
        mount_string =  '{}:{}'.format(os.path.abspath(workDir), '/mnt')
        if mount_list:
            for mount_dir in mount_list:
                mount_string = '{},{}:{}'.format(mount_string, os.path.abspath(mount_dir), '/mnt/{}'.format(os.path.basename(mount_dir)))
        
        baseSingularityCall += ['-B', mount_string, '--pwd', '/mnt']
        
    # Problem: Multiple Singularity downloads sharing the same cache directory will
    # not work correctly. See https://github.com/sylabs/singularity/issues/3634
    # and https://github.com/sylabs/singularity/issues/4555.
    
    # As a workaround, we have out own cache which we manage ourselves.
    cache_dir = os.path.join(os.environ.get('SINGULARITY_CACHEDIR',  os.path.join(os.environ.get('HOME'), '.singularity')), 'toil')
    os.makedirs(cache_dir, exist_ok=True)
    
    # What Singularity url/spec do we want?
    source_image = _convertImageSpec(tool) 
    
    # What name in the cache dir do we want?
    # We cache everything as sandbox directories and not .sif files because, as
    # laid out in https://github.com/sylabs/singularity/issues/4617, there
    # isn't a way to run from a .sif file and have write permissions on system
    # directories in the container, because the .sif build process makes
    # everything owned by root inside the image. Since some toil-vg containers
    # (like the R one) want to touch system files (to install R packages at
    # runtime), we do it this way to act more like Docker.
    #
    # Also, only sandbox directories work with user namespaces, and only user
    # namespaces work inside unprivileged Docker containers like the Toil
    # appliance.
    sandbox_dirname = os.path.join(cache_dir, '{}.sandbox'.format(hashlib.sha256(source_image.encode()).hexdigest()))
    
    if not os.path.exists(sandbox_dirname):
        # We atomically drop the sandbox at that name when we get it
        
        # Make a temp directory to be the sandbox
        temp_sandbox_dirname = tempfile.mkdtemp(dir=cache_dir)

        # Download with a fresh cache to a sandbox
        download_env = os.environ.copy()
        download_env['SINGULARITY_CACHEDIR'] = job.fileStore.getLocalTempDir()
        subprocess.check_call(['singularity', 'build', '-s', '-F', temp_sandbox_dirname, source_image], env=download_env)
        
        # Clean up the Singularity cache since it is single use
        shutil.rmtree(download_env['SINGULARITY_CACHEDIR'])
        
        try:
            # This may happen repeatedly but it is atomic
            os.rename(temp_sandbox_dirname, sandbox_dirname)
        except (FileExistsError, OSError) as e:
            # Can't rename a directory over another
            # Make sure someone else has made the directory
            assert os.path.exists(sandbox_dirname)
            # Remove our redundant copy
            shutil.rmtree(temp_sandbox_dirname)
            
        # TODO: we could save some downloading by having one process download
        # and the others wait, but then we would need a real fnctl locking
        # system here.

    # Make subprocess call for singularity run
    
    # Set the TMPDIR environment variable to relative path '.' when running an already built
    # container. This is to get around issues with differences between TMPDIR path accessibility
    # when running Singularity build and singularity exec
    download_env = os.environ.copy()
    if not 'rocker/tidyverse' in tool: 
        download_env['TMPDIR'] = '.'
    
    # If parameters is list of lists, treat each list as separate command and chain with pipes
    if len(parameters) > 0 and type(parameters[0]) is list:
        # When piping, all arguments now get merged into a single string to bash -c.
        # We try to support spaces in paths by wrapping them all in quotes first.
        chain_params = [' '.join(p) for p in [list(map(pipes.quote, q)) for q in parameters]]
        # Use bash's set -eo pipefail to detect and abort on a failure in any command in the chain
        call = baseSingularityCall + [sandbox_dirname, '/bin/bash', '-c',
                                 'set -eo pipefail && {}'.format(' | '.join(chain_params))]
    else:
        call = baseSingularityCall + [sandbox_dirname] + parameters
    logger.info("Calling singularity with " + repr(call))

    params = {}
    params['env'] = download_env
    if outfile:
        params['stdout'] = outfile
    if checkOutput:
        callMethod = subprocess.check_output
    else:
        callMethod = subprocess.check_call

    out = callMethod(call, **params)

    # After Singularity exits, it is possible that cleanup of the container's
    # temporary files is still in progress (sicne it also waits for the
    # container to exit). If we return immediately and the Toil job then
    # immediately finishes, we can have a race between Toil's temp
    # cleanup/space tracking code and Singularity's temp cleanup code to delete
    # the same directory tree. Toil doesn't handle this well, and crashes when
    # files it expected to be able to see are missing (at least on some
    # versions). So we introduce a delay here to try and make sure that
    # Singularity wins the race with high probability.
    #
    # See https://github.com/sylabs/singularity/issues/1255
    time.sleep(0.5)

    return out
    
def _convertImageSpec(spec):
    """
    Given an image specifier that may be either a Docker container specifier,
    or a Singularity URL or filename, produce the Singularity URL or filename
    that points to it.
    
    This consists of identifying the Docker container specifiers and prefixing
    them with "docker://".
    """
   
    if spec.startswith('/'):
        # It's a file path we can use.
        # Relative paths won't work because Toil uses unique working
        # directories.
        return spec
   
    if '://' in spec:
        # Already a URL
        return spec
    
    # Try it as a Docker specifier
    return 'docker://' + spec
    
    



#!/usr/bin/env python2.7
"""
vg_chunk.py: split a graph and/or GAM into chunks by connected component

"""

import argparse, sys, os, os.path, errno, random, subprocess, shutil, itertools, glob, tarfile
import doctest, re, json, collections, time, timeit
import logging, logging.handlers, struct, socket, threading
import string
import getpass
import pdb
import gzip
import logging

from math import ceil
from subprocess import Popen, PIPE

from toil.common import Toil
from toil.job import Job
from toil.realtimeLogger import RealtimeLogger
from toil_vg.vg_common import *
from toil_vg.context import Context, run_write_info_to_outstore

logger = logging.getLogger(__name__)

def chunk_subparser(parser):
    """
    Create a subparser for augmenting.  Should pass in results of subparsers.add_parser()
    """

    # Add the Toil options so the job store is the first argument
    Job.Runner.addToilOptions(parser)
    
    # General options    
    parser.add_argument("out_store",
                        help="output store.  All output written here. Path specified using same syntax as toil jobStore")
    parser.add_argument("--gam", type=make_url,
                        help="GAM to chunk")
    parser.add_argument("--graph", type=make_url, required=True,
                        help="graph or xg index to chunk")
        
    # Add common options shared with everybody
    add_common_vg_parse_args(parser)

    # Add augment options
    chunk_parse_args(parser)

    # Add common docker options
    add_container_tool_parse_args(parser)


def chunk_parse_args(parser, path_components=True):
    """
    Define chunk arguments that may be shared with other commands
    """

    parser.add_argument("--connected_component_chunking", action="store_true",
                        help="split into connected components")
    parser.add_argument("--ref_path_chunking", action="store_true",
                        help="chunk on --ref_paths if specified, all paths otherwise")
    parser.add_argument("--ref_paths", nargs='+', default=[],
                        help="reference paths to call (and chunk) on")    
    parser.add_argument("--output_format", choices=["pg", "hg", "vg"], default="pg",
                        help="output format [pg]")
    parser.add_argument("--chunk_cores", type=int,
                        help="number of threads used for extracting chunks for calling")
    parser.add_argument("--chunk_mem", type=str,
                        help="memory alotment for extracting chunks for calling")

    
def validate_chunk_options(options, chunk_optional=False):
    num_opts = [options.connected_component_chunking, options.ref_path_chunking].count(True)
    if chunk_optional == False:
        require(num_opts == 1,
                "Must specify (exactly) one of --connected_component_chunking, ref_path_chunking")
        if options.ref_paths:
            require(options.ref_path_chunking == True,
                "Must specify --ref_path_chunking when using --ref_paths")
    else:
        require(num_opts in [0, 1],
                "Must specify at most one of --connected_component_chunking or --ref_path_chunking")

    if options.gam:
        require(options.gam.endswith('.gam'),
                "Input GAM file must have .gam extension")

def run_chunking(job, context,
                 graph_id,
                 graph_basename,
                 chunk_paths,
                 connected_component_chunking,
                 output_format,
                 gam_id = None,
                 to_outstore = False):
    """ thin wrapper of vg chunk.  it will return a map from path name to id of chunked file"""

    work_dir = job.fileStore.getLocalTempDir()

    # Read our input files from the store
    graph_path = os.path.join(work_dir, graph_basename)
    job.fileStore.readGlobalFile(graph_id, graph_path)
    input_opts = ['-x', os.path.basename(graph_path)]
    if gam_id:
        gam_path = os.path.join(work_dir, 'aln.gam')
        job.fileStore.readGlobalFile(gam_id, gam_path)
        input_opts += ['-a', os.path.basename(gam_path), '-g']

    paths_path = os.path.join(work_dir, 'paths.txt')        
    if chunk_paths:
        with open(paths_path, 'w') as path_file:
            for chunk_path in chunk_paths:
                path_file.write(chunk_path + '\n')
        input_opts += ['-P', os.path.basename(paths_path)]

    # output options
    chunk_prefix = 'chunk/{}'.format(os.path.splitext(graph_basename)[0])
    os.makedirs(os.path.join(work_dir, os.path.dirname(chunk_prefix)))    
    output_opts = ['-b', chunk_prefix, ]
    output_bed_path = os.path.join(work_dir, 'chunks.bed')
    output_opts += ['-E', os.path.basename(output_bed_path)]
    output_opts += ['-O', output_format]

    # general options
    if connected_component_chunking or len(chunk_paths) > 0:
        gen_opts = ['-C']
    else:
        gen_opts = ['-M']
    gen_opts += ['-t', str(job.cores)]

    # Run vg chunk
    try:
        context.runner.call(job, ['vg', 'chunk'] + gen_opts + input_opts + output_opts,
                            work_dir = work_dir)
    except Exception as e:
        logging.error("Chunk failed. Dumping input files to outstore.")
        for dump_path in [graph_path, gam_path, paths_path]:
            if dump_path and os.path.isfile(dump_path):
                context.write_output_file(job, dump_path)
        raise
        
    # Scrape the BED into dictionary that maps path name to file id
    chunk_output = {}
    write_fn = context.write_output_file if to_outstore else context.write_intermediate_file
    with open(output_bed_path) as output_bed:
        for line in output_bed:
            toks = line.split('\t')
            if len(toks) > 3:
                graph_chunk_path = os.path.join(work_dir, toks[3].rstrip())
                # be robust to the vagaries of vg chunk: deal with graph or gam extension in bed
                if graph_chunk_path.endswith('.gam'):
                    graph_chunk_path = remove_ext(graph_chunk_path, '.gam') + '.' + output_format
                graph_chunk_id = write_fn(job, graph_chunk_path)
                chunk_output[toks[0]] = [graph_chunk_id, os.path.basename(graph_chunk_path)]
                if gam_id:
                    gam_chunk_path = os.path.join(work_dir, toks[3].rstrip())
                    if not gam_chunk_path.endswith('.gam'):
                        gam_chunk_path = remove_ext(gam_chunk_path, '.' + output_format) + '.gam'
                    # vg chunk's not going to write empty files, so make sure we have one
                    open(gam_chunk_path, 'a').close()
                    gam_chunk_id = write_fn(job, gam_chunk_path)
                    chunk_output[toks[0]] += [gam_chunk_id, os.path.basename(gam_chunk_path)]

    return chunk_output
    
    
def chunk_main(context, options):
    """ entrypoint for calling """

    validate_chunk_options(options)
            
    # How long did it take to run the entire pipeline, in seconds?
    run_time_pipeline = None
        
    # Mark when we start the pipeline
    start_time_pipeline = timeit.default_timer()
    
    with context.get_toil(options.jobStore) as toil:
        if not toil.options.restart:

            importer = AsyncImporter(toil)

            # Upload local files to the job store
            inputGraphFileID = importer.load(options.graph)
            inputGamFileID = None
            if options.gam:
                inputGamFileID = importer.load(options.gam)

            importer.wait()

            # Make a root job
            root_job = Job.wrapJobFn(run_chunking, context,
                                     importer.resolve(inputGraphFileID),
                                     os.path.basename(options.graph),
                                     chunk_paths=options.ref_paths,
                                     connected_component_chunking=options.connected_component_chunking,
                                     output_format=options.output_format,
                                     gam_id = importer.resolve(inputGamFileID),
                                     to_outstore = True,
                                     cores=context.config.chunk_cores,
                                     memory=context.config.chunk_mem,
                                     disk=context.config.chunk_disk)

            # Init the outstore
            init_job = Job.wrapJobFn(run_write_info_to_outstore, context, sys.argv,
                                     memory=context.config.misc_mem,
                                     disk=context.config.misc_disk)
            init_job.addFollowOn(root_job)            
            
            # Run the job and store the returned list of output files to download
            toil.start(init_job)
        else:
            toil.restart()
                
    end_time_pipeline = timeit.default_timer()
    run_time_pipeline = end_time_pipeline - start_time_pipeline
 
    logger.info("All jobs completed successfully. Pipeline took {} seconds.".format(run_time_pipeline))
    
    



#!/usr/bin/env python2.7
"""
vg_augment.py: augment a vg graph to include variation from a GAM alignment

"""

import argparse, sys, os, os.path, errno, random, subprocess, shutil, itertools, glob, tarfile
import doctest, re, json, collections, time, timeit
import logging, logging.handlers, struct, socket, threading
import string
import getpass
import pdb
import gzip
import logging

from math import ceil
from subprocess import Popen, PIPE

from toil.common import Toil
from toil.job import Job
from toil.realtimeLogger import RealtimeLogger
from toil_vg.vg_common import *
from toil_vg.context import Context, run_write_info_to_outstore
from toil_vg.vg_chunk import *

logger = logging.getLogger(__name__)

def augment_subparser(parser):
    """
    Create a subparser for augmenting.  Should pass in results of subparsers.add_parser()
    """

    # Add the Toil options so the job store is the first argument
    Job.Runner.addToilOptions(parser)
    
    # General options
    
    parser.add_argument("out_store",
                        help="output store.  All output written here. Path specified using same syntax as toil jobStore")
    parser.add_argument("--gam", type=make_url, required=True,
                        help="GAM to augment")
    parser.add_argument("--graph", type=make_url, required=True,
                        help="graph to augment")
        
    # Add common options shared with everybody
    add_common_vg_parse_args(parser)

    # Add augment options
    augment_parse_args(parser, stand_alone = True)

    # Add common chunking options shared with vg_chunk
    chunk_parse_args(parser)

    # Add common docker options
    add_container_tool_parse_args(parser)


def augment_parse_args(parser, stand_alone = False):
    """
    Define map arguments shared with mapeval and run
    """
    if stand_alone:
        parser.add_argument("--augment_gam", action="store_true",
                            help="produce and augmented GAM")    
    parser.add_argument("--min_augment_coverage", type=int, 
                        help="minimum coverage for breakpoint to be applied")
    parser.add_argument("--expected_coverage", type=int,
                        help="expected coverage.  only affects memory usage.  use if coverage >> 100")
    parser.add_argument("--min_mapq", type=int,
                        help="ignore reads with MAPQ less than this")
    parser.add_argument("--min_baseq", type=int,
                        help="ignore edits with minimum average base quality less than this")
    parser.add_argument("--augment_cores", type=int,
                        help="number of threads during augmentation")
    parser.add_argument("--augment_mem", type=str,
                        help="memory alotment during augmentation")

def run_chunked_augmenting(job, context,
                           graph_id,
                           graph_basename,
                           gam_id,
                           gam_basename,
                           batch_input=None,
                           all_path_components=False,
                           chunk_paths=[],
                           connected_component_chunking=False,
                           output_format=None,
                           augment_gam=False,
                           min_augment_coverage=None,
                           expected_coverage=None,
                           min_mapq=None, 
                           min_baseq=None,
                           to_outstore=False):
    """
    Run a chunking job (if desired), then augment the results
    """

    # base case: only one input
    if batch_input is None:
        # chunk if necessary
        if all_path_components or connected_component_chunking or len(chunk_paths) > 1:
            child_job = Job()
            job.addChild(child_job)
            chunk_job = child_job.addChildJobFn(run_chunking, context,
                                                graph_id=graph_id,
                                                graph_basename=graph_basename,
                                                chunk_paths=chunk_paths,
                                                connected_component_chunking=connected_component_chunking,
                                                output_format=output_format,
                                                gam_id=gam_id,
                                                to_outstore=False,
                                                cores=context.config.chunk_cores,
                                                memory=context.config.chunk_mem,
                                                disk=context.config.chunk_disk)
            batch_input = chunk_job.rv()

            # recurse on chunks
            recurse_job = child_job.addFollowOnJobFn(run_chunked_augmenting, context,
                                                     graph_id=None,
                                                     graph_basename=None,
                                                     gam_id=None,
                                                     gam_basename=None,
                                                     batch_input=batch_input,
                                                     all_path_components=all_path_components,
                                                     chunk_paths=chunk_paths,
                                                     connected_component_chunking=connected_component_chunking,
                                                     output_format=output_format,
                                                     augment_gam=augment_gam,
                                                     min_augment_coverage=min_augment_coverage,
                                                     expected_coverage=expected_coverage,
                                                     min_mapq=min_mapq, 
                                                     min_baseq=min_baseq,
                                                     to_outstore=to_outstore)
            return recurse_job.rv()
        else:
            #phony up chunk output for single input
            batch_input = { 'all' : [graph_id, graph_basename] }
            if gam_id:
                batch_input['all'] += [gam_id, gam_basename]

    # run the augmenting on each chunk
    assert batch_input

    augment_results = []
    for chunk_name, chunk_results in list(batch_input.items()):
        augment_job = job.addChildJobFn(run_augmenting, context,
                                        graph_id=chunk_results[0],
                                        graph_basename=chunk_results[1],
                                        gam_id=chunk_results[2],
                                        gam_basename=chunk_results[3],
                                        augment_gam=augment_gam,
                                        min_augment_coverage=min_augment_coverage,
                                        expected_coverage=expected_coverage,
                                        min_mapq=min_mapq,
                                        min_baseq=min_baseq,
                                        to_outstore=to_outstore,
                                        cores=context.config.augment_cores,
                                        memory=context.config.augment_mem,
                                        disk=context.config.augment_disk)
        
        augment_results.append((chunk_name, augment_job.rv()))

    return augment_results
        

def run_augmenting(job, context,
                  graph_id,
                  graph_basename,
                  gam_id,
                  gam_basename,
                  augment_gam=False,
                  min_augment_coverage=None,
                  expected_coverage=None,
                  min_mapq=None,
                  min_baseq=None,
                  to_outstore=False):
    """
    Augment the graph (and gam if wanted)
    """

    work_dir = job.fileStore.getLocalTempDir()

    # Read our input files from the store
    graph_path = os.path.join(work_dir, graph_basename)
    job.fileStore.readGlobalFile(graph_id, graph_path)    
    gam_path = os.path.join(work_dir, gam_basename)
    job.fileStore.readGlobalFile(gam_id, gam_path)

    augment_cmd = ['vg', 'augment', os.path.basename(graph_path), os.path.basename(gam_path), '-t', str(job.cores)]

    # hardcoded naming convention: tack on an -aug to input paths
    graph_name, graph_ext = os.path.splitext(graph_basename)
    augmented_graph_path = os.path.join(work_dir, graph_name + '-aug' + graph_ext)
    augmented_gam_path = os.path.join(work_dir, remove_ext(gam_basename, '.gam') + '-aug.gam')
    if augment_gam:
        augment_cmd += ['-A', os.path.basename(augmented_gam_path)]

    # optional stuff
    if min_augment_coverage is not None:
        augment_cmd += ['-m', str(min_augment_coverage)]
    if expected_coverage is not None:
        augment_cmd += ['-c', str(expected_coverage)]
    if min_mapq is not None:
        augment_cmd += ['-Q', str(min_mapq)]
    if min_baseq is not None:
        augment_cmd += ['-q', str(min_baseq)]
    if context.config.augment_opts:
        augment_cmd += context.config.augment_opts
    # always support subgraphs
    augment_cmd += ['-s']

    # run the command
    try:
        with open(augmented_graph_path, 'wb') as augmented_graph_file:
            context.runner.call(job, augment_cmd, work_dir = work_dir, outfile = augmented_graph_file)
    except Exception as e:
        logging.error("Augment failed. Dumping input files to outstore.")
        for dump_path in [graph_path, gam_path]:
            if dump_path and os.path.isfile(dump_path):
                context.write_output_file(job, dump_path)
        raise

    # return the output
    write_fn = context.write_output_file if to_outstore else context.write_intermediate_file
    aug_graph_id = write_fn(job, augmented_graph_path)
    aug_gam_id = write_fn(job, augmented_gam_path) if augment_gam else None
    return aug_graph_id, aug_gam_id
    
def augment_main(context, options):
    """
    Wrapper for vg augment.
    """

    validate_chunk_options(options, chunk_optional=True)
        
    # How long did it take to run the entire pipeline, in seconds?
    run_time_pipeline = None
        
    # Mark when we start the pipeline
    start_time_pipeline = timeit.default_timer()

    with context.get_toil(options.jobStore) as toil:
        if not toil.options.restart:

            importer = AsyncImporter(toil)

            # Upload local files to the job store
            inputGraphFileID = importer.load(options.graph)
            inputGamFileID = importer.load(options.gam)

            importer.wait()

            # Make a root job
            root_job = Job.wrapJobFn(run_chunked_augmenting, context,
                                     graph_id = importer.resolve(inputGraphFileID),
                                     graph_basename = os.path.basename(options.graph),
                                     gam_id = importer.resolve(inputGamFileID),
                                     gam_basename = os.path.basename(options.gam),
                                     batch_input=None,
                                     all_path_components=options.ref_path_chunking,
                                     chunk_paths=options.ref_paths,
                                     connected_component_chunking=options.connected_component_chunking,
                                     output_format=options.output_format,
                                     augment_gam=options.augment_gam,
                                     min_augment_coverage=options.min_augment_coverage,
                                     expected_coverage=options.expected_coverage,
                                     min_mapq=options.min_mapq,
                                     min_baseq=options.min_baseq,
                                     to_outstore=True)

            # Init the outstore
            init_job = Job.wrapJobFn(run_write_info_to_outstore, context, sys.argv,
                                     memory=context.config.misc_mem,
                                     disk=context.config.misc_disk)
            init_job.addFollowOn(root_job)            
            
            # Run the job and store the returned list of output files to download
            toil.start(init_job)
        else:
            toil.restart()

    end_time_pipeline = timeit.default_timer()
    run_time_pipeline = end_time_pipeline - start_time_pipeline
 
    logger.info("All jobs completed successfully. Pipeline took {} seconds.".format(run_time_pipeline))



#!/usr/bin/env python
"""
pedigree_analysis.py: run the UDP analysis workflow on a pedigree processed by the 'toil-vg pedigree' subcommand workflow.

"""

import argparse, sys, os, os.path, errno, random, subprocess, shutil, itertools, glob, tarfile
import doctest, re, json, collections, time, timeit
import logging, logging.handlers, struct, socket, threading
import string
import getpass
import pdb
import gzip
import logging

from math import ceil
from subprocess import Popen, PIPE

from toil.common import Toil
from toil.job import Job
from toil.realtimeLogger import RealtimeLogger
from toil_vg.vg_common import *
from toil_vg.context import Context, run_write_info_to_outstore

logger = logging.getLogger(__name__)

def analysis_subparser(parser):
    """
    Create a subparser for the analysis workflow.  Should pass in results of subparsers.add_parser()
    """

    # Add the Toil options so the job store is the first argument
    Job.Runner.addToilOptions(parser)
    
    # General options
    parser.add_argument("out_store",
        help="output store.  All output written here. Path specified using same syntax as toil jobStore")
    # VCFtoShebang options
    parser.add_argument("--cohort_vcf", type=make_url, required=True,
        help="Path to cohort vcf processed by toil-vg pedigree")    
    parser.add_argument("--sample_name", type=str, required=True,
        help="Cohort proband sample name as present in the sample column of --cohort_vcf")
    parser.add_argument("--bypass", action="store_true", default=False,
        help="Parameter for vcftoshebang.")
    # TODO: what is the bypass parameter?
    parser.add_argument("--cadd_lines", type=int, default=4985,
        help="Parameter for vcftoshebang.")
    # TODO: what is the cadd lines parameter?
    parser.add_argument("--chrom_dir", type=str, required=True,
        help="Path to chromosome annotation directory used by vcftoshebang")
    parser.add_argument("--edit_dir", type=str, required=True,
        help="Path to directory containing master edit files used by vcftoshebang")
    
    # CADD options
    parser.add_argument("--split_lines", type=int, default=30000,
        help="Number of lines to chunk the input VCF for CADD processing.")
    parser.add_argument("--genome_build", type=str, default="GRCh37",
        help="Genome annotation version for the CADD engine")
    parser.add_argument("--cadd_data", type=str, required=True,
        help="Path to cadd engine data directory")
    
    # BMTB options
    parser.add_argument("--maternal_bam", type=make_url, required=True,
        help="Input maternal bam file")    
    parser.add_argument("--maternal_bai", type=make_url, required=True,
        help="Input maternal bam file .bai index")    
    parser.add_argument("--paternal_bam", type=make_url, required=True,
        help="Input paternal bam file")    
    parser.add_argument("--paternal_bai", type=make_url, required=True,
        help="Input paternal bam file .bai index")    
    parser.add_argument("--siblings_bam", nargs='+', type=make_url, required=True,
        help="INPUT sibling bam files. Proband bam file must be listed 1st.")
    parser.add_argument("--siblings_bai", nargs='+', type=make_url, required=True,
        help="INPUT sibling bam file .bai indicies. Same sample order as --siblings_bam.")
    parser.add_argument("--maternal_name", type=str, required=True,
        help="Sample name of the pedigree mother.")
    parser.add_argument("--paternal_name", type=str, required=True,
        help="Sample name of the pedigree father.")
    parser.add_argument("--sibling_names", nargs='+', type=str, required=True,
        help="Sample names of the siblings. Same sample order as --siblings_bam.")
    parser.add_argument("--sibling_genders", nargs='+', type=int, required=True,
        help="Gender of each sibling sample. 0 = male, 1 = female. Same sample order as --siblings_bam.")
    parser.add_argument("--sibling_affected", nargs='+', type=int, required=True,
        help="Affected status of each sibling sample. 0 = unaffected, 1 = affected. Same sample order as --siblings_bam.")
    
    # Add common options shared with everybody
    add_common_vg_parse_args(parser)

    # Add common docker options
    add_container_tool_parse_args(parser)

def run_vcftoshebang(job, context, sample_name, cohort_vcf_id, bypass, cadd_lines, chrom_dir, edit_dir):
    """ run vcftoshebang on a cohort vcf.
    
    Takes a joint-called cohort vcf from the vg_pedigree.py workflow along with the sample name of the proband,
    the bypass boolean parameter for vcftoshebang, the number of lines to split the varsifter variant list that
    will be passed to the CADD engine, the full path to the directory containing chromosome config files, and
    the full path to the directory containing the master edit files.
    
    Returns a tuple containing varsifter formatted and annotated version of the cohort vcf as the first element, and
    a list of file ids that will passed to the CADD engine if any exist.
    
    """
    # Define work directory for docker calls
    work_dir = job.fileStore.getLocalTempDir()

    cohort_vcf_file = os.path.join(work_dir, os.path.basename(cohort_vcf_id))
    job.fileStore.readGlobalFile(cohort_vcf_id, cohort_vcf_file)
    
    # Decompress cohort vcf if already compressed
    if os.path.basename(cohort_vcf_file).endswith('.gz'):
        context.runner.call(job, ['bgzip', '-d', os.path.basename(cohort_vcf_file)], work_dir = work_dir, tool_name='vg')
        cohort_vcf_file = os.path.splitext(cohort_vcf_file)[0]
    
    # Remove the hs37d5 contig
    context.runner.call(job, ['vcftools', '--vcf', os.path.basename(cohort_vcf_file), '--not-chr', 'hs37d5', '--recode-INFO-all', '--recode', '--out', '{}.filtered'.format(os.path.basename(os.path.splitext(cohort_vcf_file)[0]))], work_dir = work_dir, tool_name='vcftools')
    input_vcf_file = "{}.filtered.recode.vcf".format(os.path.basename(os.path.splitext(cohort_vcf_file)[0]))
    
    output_dir = "vcf2shebang_output/"
    bypass_conf = "NO"
    if bypass == 'true': bypass_conf = "YES"
    context.runner.call(job, ['mkdir', '{}'.format(output_dir)], work_dir = work_dir)
    cmd_list = []
    cmd_list.append(['cp', '/vcftoshebang/VCFtoShebang_Config.txt', '.'])
    cmd_list.append(['sed', '-i', '\"s|.*PROBAND_NAME.*|PROBAND_NAME\t{}|\"'.format(sample_name), 'VCFtoShebang_Config.txt'])
    cmd_list.append(['sed', '-i', '\"s|.*OUTPUT_DIR.*|OUTPUT_DIR\t{}|\"'.format(output_dir), 'VCFtoShebang_Config.txt'])
    cmd_list.append(['sed', '-i', '\"s|.*UNROLLED_VCF_PATH.*|UNROLLED_VCF_PATH\t{}|\"'.format(input_vcf_file), 'VCFtoShebang_Config.txt'])
    cmd_list.append(['sed', '-i', '\"s|.*BYPASS.*|BYPASS\t{}|\"'.format(bypass_conf), 'VCFtoShebang_Config.txt'])
    cmd_list.append(['sed', '-i', '\"s|.*CADD_LINES.*|CADD_LINES\t{}|\"'.format(cadd_lines), 'VCFtoShebang_Config.txt'])
    cmd_list.append(['sed', '-i', '\"s|.*CHROM_DIR.*|CHROM_DIR\t$PWD/{}|\"'.format(os.path.basename(os.path.normpath(chrom_dir))), 'VCFtoShebang_Config.txt'])
    cmd_list.append(['sed', '-i', '\"s|.*EDIT_DIR.*|EDIT_DIR\t$PWD/{}|\"'.format(os.path.basename(os.path.normpath(edit_dir))), 'VCFtoShebang_Config.txt'])
    cmd_list.append(['sed', '-i', '\"s|.*EDITOR_CONFIG.*|EDITOR_CONFIG\t/vcftoshebang/edit_config.txt|\"', 'VCFtoShebang_Config.txt'])
    cmd_list.append(['java', '-XX:+UnlockExperimentalVMOptions', '-XX:ActiveProcessorCount=32', '-cp', '/vcftoshebang/VCFtoShebang.jar:/vcftoshebang/json_simple.jar',
                             'Runner', 'VCFtoShebang_Config.txt'])
    chain_cmds = [' '.join(p) for p in cmd_list]
    command = ['/bin/bash', '-c', 'set -eo pipefail && {}'.format(' && '.join(chain_cmds))]
    context.runner.call(job, command, work_dir = work_dir, tool_name='vcf2shebang', mount_list=[chrom_dir,edit_dir])
    is_empty = True
    with open(os.path.join(work_dir, 'vcf2shebang_output/{}_unrolled_snpeff_fix_overlap_mono_CADD_Input_Files/{}_unrolled_snpeff_fix_overlap_mono_CADD_input_file.txt'.format(sample_name,sample_name)), 'r') as cadd_input_file:
        for line in cadd_input_file:
            if "#" in line: continue
            else:
                is_empty = False
                break
    
    output_vs_path = os.path.join(work_dir, 'vcf2shebang_output/{}_unrolled_snpeff_fix_overlap_mono_shebang.vs'.format(sample_name))
    output_cadd_vcf_path = None
    if not is_empty:
        output_cadd_vcf_path = os.path.join(work_dir, 'vcf2shebang_output/{}_unrolled_snpeff_fix_overlap_mono_CADD_Input_Files/{}_unrolled_snpeff_fix_overlap_mono_CADD_input_file.txt.gz'.format(sample_name,sample_name))
    
    return context.write_output_file(job, output_vs_path), context.write_output_file(job, output_cadd_vcf_path)
    
def run_split_vcf(job, context, vcf_file_id, split_lines):
    """ split vcf into chunks for passing through to the CADD engine for CADD scoring.
    """
    # Define work directory for docker calls
    work_dir = job.fileStore.getLocalTempDir()

    vcf_file = os.path.join(work_dir, os.path.basename(vcf_file_id))
    job.fileStore.readGlobalFile(vcf_file_id, vcf_file)
    
    tempname = "{}_tmp".format(os.path.basename(os.path.splitext(os.path.splitext(vcf_file)[0])[0]))
    cmd = [['zcat', os.path.basename(vcf_file)]]
    cmd.append(['grep', '-v', '^GL|^#'])
    cmd.append(['cut', '-f', '1-5'])
    cmd.append(['split', '-l', str(split_lines), '--additional-suffix=.vcf', '-d', '-', tempname])
    context.runner.call(job, cmd, work_dir = work_dir)
    
    vcf_chunk_ids = []
    for chunk_name in sorted(os.listdir(work_dir)):
        if chunk_name.endswith('.vcf') and 'tmp' in chunk_name:
            vcf_chunk_ids.append(context.write_intermediate_file(job, os.path.join(work_dir, chunk_name)))
    
    return vcf_chunk_ids

def run_cadd(job, context, chunk_vcf_id, genome_build, cadd_data_dir):
    """ run the main CADD engine on vcf chunks.
    """
    # Define work directory for docker calls
    work_dir = job.fileStore.getLocalTempDir()

    vcf_file = os.path.join(work_dir, os.path.basename(chunk_vcf_id))
    job.fileStore.readGlobalFile(chunk_vcf_id, vcf_file)
    
    base_vcf_name = os.path.basename(os.path.splitext(vcf_file)[0])
    cadd_data_dir_basename = os.path.basename(cadd_data_dir)
    cmd_list = []
    cmd_list.append(['source', 'activate', '$(head', '-1', '/usr/src/app/environment.yml', '|', 'cut', '-d\'', '\'', '-f2)'])
    cmd_list.append(['/bin/bash', '/usr/src/app/CADD.sh', '-v', '\"v1.5\"', '-g', genome_build, '-o', '$PWD/{}_out.tsv.gz'.format(base_vcf_name), '-d', '$PWD/{}'.format(cadd_data_dir_basename), os.path.basename(vcf_file)])
    chain_cmds = [' '.join(p) for p in cmd_list]
    command = ['/bin/bash', '-c', 'set -eo pipefail && {}'.format(' && '.join(chain_cmds))]
    context.runner.call(job, command, work_dir = work_dir, tool_name='cadd', mount_list=[cadd_data_dir])
    
    output_cadd_path = os.path.join(work_dir, '{}_out.tsv.gz'.format(base_vcf_name))
    return context.write_output_file(job, output_cadd_path)

def run_merge_annotated_vcf(job, context, cadd_output_chunk_ids):
    """ run the merge task for the CADD engine output chunks.
    """
    # Define work directory for docker calls
    work_dir = job.fileStore.getLocalTempDir()
    
    cmd_list = [['source', 'activate', '$(head', '-1', '/usr/src/app/environment.yml', '|', 'cut', '-d\' \'', '-f2)']]
    # Decompress and concatenate cadd output chunks into a single file
    cadd_chunk_merged_file = os.path.join(work_dir, 'merged_CADDv1.5_offline_unsorted')
    for cadd_output_chunk_id in sorted(cadd_output_chunk_ids):
        cadd_chunk_file = os.path.join(work_dir, os.path.basename(cadd_output_chunk_id))
        job.fileStore.readGlobalFile(cadd_output_chunk_id, cadd_chunk_file)
        cmd_list.append(['zcat', os.path.basename(cadd_chunk_file), '>>', 'merged_CADDv1.5_offline_unsorted'])
        
    # Sort and process the merged file
    cmd_list.append(['sort', '-k1,1', '-k2,2n', 'merged_CADDv1.5_offline_unsorted', '>', 'merged_CADDv1.5_offline.vcf'])
    cmd_list.append(['rm', '-f', 'merged_CADDv1.5_offline_unsorted'])
    cmd_list.append(['python', '/usr/src/app/CADD_offline_mito_postprocessing.py', '-c', '/usr/src/app/whole_mito_SNP_pp2_predictions_sorted.txt', '-i', 'merged_CADDv1.5_offline.vcf', '-o', 'merged_CADDv1.5_offline_proper_format.vcf'])
    cmd_list.append(['rm', '-f', 'merged_CADDv1.5_offline.vcf'])
    chain_cmds = [' '.join(p) for p in cmd_list]
    command = ['/bin/bash', '-c', 'set -eo pipefail && {}'.format(' && '.join(chain_cmds))]
    context.runner.call(job, command, work_dir = work_dir, tool_name='cadd')
    
    merged_cadd_output_vcf_path = os.path.join(work_dir, 'merged_CADDv1.5_offline_proper_format.vcf')
    return context.write_output_file(job, merged_cadd_output_vcf_path)

def run_cadd_editor(job, context, vcftoshebang_vs_file_id, merged_cadd_vcf_file_id):
    """ run editor for merged CADD output postprocessing and merges annotations with vcftoshebang varsifter output.
        Outputs the file in Varsifter format.
    """
    # Define work directory for docker calls
    work_dir = job.fileStore.getLocalTempDir()
    
    vcftoshebang_vs_file = os.path.join(work_dir, os.path.basename(vcftoshebang_vs_file_id))
    job.fileStore.readGlobalFile(vcftoshebang_vs_file_id, vcftoshebang_vs_file)
    
    merged_cadd_vcf_file = os.path.join(work_dir, os.path.basename(merged_cadd_vcf_file_id))
    job.fileStore.readGlobalFile(merged_cadd_vcf_file_id, merged_cadd_vcf_file)
    
    command = ['java', '-cp', '/cadd_edit/NewCaddEditor.jar:/cadd_edit/commons-cli-1.4.jar', 'NewCaddEditor',
                '--input_vs', os.path.basename(vcftoshebang_vs_file), '--output_cadd', os.path.basename(merged_cadd_vcf_file), '--output_vs', 'cadd_editor_output.vs']
    context.runner.call(job, command, work_dir = work_dir, tool_name='caddeditor')
    
    cadd_editor_output_path = os.path.join(work_dir, 'cadd_editor_output.vs')
    return context.write_output_file(job, cadd_editor_output_path)

def run_bmtb(job, context, analysis_ready_vs_file_id,
               maternal_bam_id, maternal_bai_id, paternal_bam_id, paternal_bai_id, sibling_bam_ids, sibling_bai_ids,
               maternal_name, paternal_name, sibling_names, sibling_genders, sibling_affected):
    """ run the Black Magic Toolbox candidate analysis program on the input varsifter file, cohort bam files, and sibling gender
        and sibling affected status.
    """
    # Define work directory for docker calls
    work_dir = job.fileStore.getLocalTempDir()
    
    vs_file_path = os.path.join(work_dir, os.path.basename(analysis_ready_vs_file_id))
    job.fileStore.readGlobalFile(analysis_ready_vs_file_id, vs_file_path)
    
    m_bam_path = os.path.join(work_dir, os.path.basename(maternal_bam_id))
    job.fileStore.readGlobalFile(maternal_bam_id, m_bam_path)
    m_bai_path = os.path.join(work_dir, os.path.basename(maternal_bai_id))
    job.fileStore.readGlobalFile(maternal_bai_id, m_bai_path)
    
    f_bam_path = os.path.join(work_dir, os.path.basename(paternal_bam_id))
    job.fileStore.readGlobalFile(paternal_bam_id, f_bam_path)
    f_bai_path = os.path.join(work_dir, os.path.basename(paternal_bai_id))
    job.fileStore.readGlobalFile(paternal_bai_id, f_bai_path)
    
    s_bam_paths = []
    for s_bam_id in sibling_bam_ids:
        s_bam_path = os.path.join(work_dir, os.path.basename(s_bam_id))
        job.fileStore.readGlobalFile(s_bam_id, s_bam_path)
        s_bam_paths.append(s_bam_path)
    s_bai_paths = []
    for s_bai_id in sibling_bai_ids:
        s_bai_path = os.path.join(work_dir, os.path.basename(s_bai_id))
        job.fileStore.readGlobalFile(s_bai_id, s_bai_path)
        s_bai_paths.append(s_bai_path)
    
    context.runner.call(job, ['ls', '-l', '/bmtb/'], work_dir = work_dir, tool_name='bmtb')
    cmd_list = [['cp', '-r', '/bmtb/Configs', '$PWD/Configs']]
    cmd_list.append(['rm', '-f', '$PWD/Configs/BAM_Directory_Config.txt'])
    cmd_list.append(['touch', '$PWD/Configs/BAM_Directory_Config.txt'])
    cmd_list.append(['echo', '-e', '\"{}\t$PWD/{}\"'.format(maternal_name,os.path.basename(m_bam_path)), '>>', '$PWD/Configs/BAM_Directory_Config.txt'])
    cmd_list.append(['echo', '-e', '\"{}\t$PWD/{}\"'.format(paternal_name,os.path.basename(f_bam_path)), '>>', '$PWD/Configs/BAM_Directory_Config.txt'])
    sibling_id_string = "NA"
    for i, (s_bam_path,s_bai_path,s_name,s_gender,s_affected) in enumerate(zip(s_bam_paths,s_bai_paths,sibling_names,sibling_genders,sibling_affected)):
        cmd_list.append(['echo', '-e', '\"{}\t$PWD/{}\"'.format(s_name,os.path.basename(s_bam_path)), '>>', '$PWD/Configs/BAM_Directory_Config.txt'])
        # Add proband data to master config file
        if i == 0:
            cmd_list.append(['sed', '-i', '\"s|.*PB_ID.*|PB_ID\t{}|\"'.format(s_name), '$PWD/Configs/BMTB_Genome_Input_Config.txt'])
            cmd_list.append(['sed', '-i', '\"s|.*PB_GENDER.*|PB_GENDER\t{}|\"'.format(s_gender), '$PWD/Configs/BMTB_Genome_Input_Config.txt'])
            cmd_list.append(['sed', '-i', '\"s|.*OUT_FILE_NAME.*|OUT_FILE_NAME\t{}|\"'.format(s_name), '$PWD/Configs/BMTB_Genome_Input_Config.txt'])
        else:
            if sibling_id_string == "NA":
                sibling_id_string = "{},{},{};".format(s_name,s_gender,s_affected)
            else:
                sibling_id_string += "{},{},{};".format(s_name,s_gender,s_affected)
    
    cmd_list.append(['sed', '-i', '\"s|.*VS_FILE_PATH.*|VS_FILE_PATH\t$PWD/{}|\"'.format(os.path.basename(vs_file_path)), '$PWD/Configs/BMTB_Genome_Input_Config.txt'])
    cmd_list.append(['sed', '-i', '\"s|.*CONFIG_FILE_DIREC.*|CONFIG_FILE_DIREC\t$PWD/Configs|\"', '$PWD/Configs/BMTB_Genome_Input_Config.txt'])
    cmd_list.append(['sed', '-i', '\"s|.*FATHER_ID.*|FATHER_ID\t{}|\"'.format(paternal_name), '$PWD/Configs/BMTB_Genome_Input_Config.txt'])
    cmd_list.append(['sed', '-i', '\"s|.*MOTHER_ID.*|MOTHER_ID\t{}|\"'.format(maternal_name), '$PWD/Configs/BMTB_Genome_Input_Config.txt'])
    cmd_list.append(['sed', '-i', '\"s|.*SIB_IDS.*|SIB_IDS\t{}|\"'.format(sibling_id_string), '$PWD/Configs/BMTB_Genome_Input_Config.txt'])
    cmd_list.append(['java', '-cp', '/bmtb/bmtb.jar:/bmtb/htsjdk-2.19.0-47-gc5ed6b7-SNAPSHOT.jar', 'general.Runner', '$PWD/Configs/BMTB_Genome_Input_Config.txt'])
    cmd_list.append(['tar', 'czvf', '\"{}_BlackBox_Output.tar.gz\"'.format(sibling_names[0]), '\"{}_BlackBox_Output\"'.format(sibling_names[0])])
    chain_cmds = [' '.join(p) for p in cmd_list]
    command = ['/bin/bash', '-c', 'set -eo pipefail && {}'.format(' && '.join(chain_cmds))]
    context.runner.call(job, command, work_dir = work_dir, tool_name='bmtb')
    
    output_package_path = os.path.join(work_dir, '{}_BlackBox_Output.tar.gz'.format(sibling_names[0]))
    return context.write_output_file(job, output_package_path)

def run_cadd_jobs(job, context, vcf_chunk_ids, genome_build, cadd_data_dir):
    """ helper function for running multiple cadd jobs per vcf chunk
    """
    cadd_engine_output_ids = []
    for vcf_chunk_id in vcf_chunk_ids:
        cadd_job = job.addChildJobFn(run_cadd, context, vcf_chunk_id, genome_build, cadd_data_dir,
                                        cores=context.config.misc_cores,
                                        memory=context.config.misc_mem,
                                        disk=context.config.misc_disk)
        cadd_engine_output_ids.append(cadd_job.rv())
    
    return cadd_engine_output_ids
    
def run_analysis(job, context, cohort_vcf_id,
                       maternal_bam_id, maternal_bai_id, paternal_bam_id, paternal_bai_id, sibling_bam_ids, sibling_bai_ids,
                       sample_name, maternal_name, paternal_name,
                       sibling_names, sibling_genders, sibling_affected,
                       bypass, cadd_lines,
                       chrom_dir, edit_dir,
                       split_lines, genome_build, cadd_data_dir):
    """ run vcf to shebang varsifter file conversion, then do cadd scoring and annotation, finally run the blackmagiktoolbox workflow.
        returns final candidate varsifter file, paired with total surject time
    (excluding toil-vg overhead such as transferring and splitting files )"""

    # to encapsulate everything under this job
    child_job = Job()
    job.addChild(child_job)
    
    vcf_to_shebang_job = child_job.addChildJobFn(run_vcftoshebang, context, sample_name, cohort_vcf_id, bypass, cadd_lines, chrom_dir, edit_dir,
                                                    cores=context.config.alignment_cores,
                                                    memory=context.config.alignment_mem,
                                                    disk=context.config.alignment_disk)
    
    analysis_ready_vs_file_id = vcf_to_shebang_job.rv(0)
    if vcf_to_shebang_job.rv(1) is not None:
        RealtimeLogger.info("Some variants don't have CADD scores, running them through the CADD engine workflow.")
        split_vcf_job = vcf_to_shebang_job.addChildJobFn(run_split_vcf, context, vcf_to_shebang_job.rv(1), cadd_lines)
        cadd_jobs = split_vcf_job.addFollowOnJobFn(run_cadd_jobs, context, split_vcf_job.rv(), genome_build, cadd_data_dir,
                                                cores=context.config.misc_cores,
                                                memory=context.config.misc_mem,
                                                disk=context.config.misc_disk)
        merge_annotated_vcf_job = cadd_jobs.addFollowOnJobFn(run_merge_annotated_vcf, context, cadd_jobs.rv())
        cadd_edit_job = merge_annotated_vcf_job.addFollowOnJobFn(run_cadd_editor, context, vcf_to_shebang_job.rv(0), merge_annotated_vcf_job.rv(),
                                                                    cores=context.config.misc_cores,
                                                                    memory=context.config.alignment_mem,
                                                                    disk=context.config.alignment_disk)
        
        analysis_ready_vs_file_id = cadd_edit_job.rv()
    
    bmtb_job = vcf_to_shebang_job.addFollowOnJobFn(run_bmtb, context, analysis_ready_vs_file_id,
                                                   maternal_bam_id, maternal_bai_id, paternal_bam_id, paternal_bai_id, sibling_bam_ids, sibling_bai_ids,
                                                   maternal_name, paternal_name, sibling_names, sibling_genders, sibling_affected,
                                                   cores=context.config.misc_cores,
                                                   memory=context.config.alignment_mem,
                                                   disk=context.config.alignment_disk)
    
    return bmtb_job.rv()
    
def analysis_main(context, options):
    """
    Wrapper for the pedigree analysis pipeline. 
    """

    # How long did it take to run the entire pipeline, in seconds?
    run_time_pipeline = None
        
    # Mark when we start the pipeline
    start_time_pipeline = timeit.default_timer()

    with context.get_toil(options.jobStore) as toil:
        if not toil.options.restart:

            importer = AsyncImporter(toil)
            
            # Upload local files to the remote IO Store
            inputVCFFileID = importer.load(options.cohort_vcf)
            inputMBAMFileID = importer.load(options.maternal_bam)
            inputMBAMINDEXFileID = importer.load(options.maternal_bai)
            inputFBAMFileID = importer.load(options.paternal_bam)
            inputFBAMINDEXFileID = importer.load(options.paternal_bai)
            inputSiblingBAMFileIDs = []
            for sibling_bam in options.siblings_bam:
                inputSiblingBAMFileIDs.append(importer.load(sibling_bam))
            inputSiblingBAMINDEXFileIDs = []
            for sibling_bai in options.siblings_bai:
                inputSiblingBAMINDEXFileIDs.append(importer.load(sibling_bai))
            
            importer.wait()

            # Make a root job
            root_job = Job.wrapJobFn(run_analysis, context,
                                     importer.resolve(inputVCFFileID),
                                     importer.resolve(inputMBAMFileID),
                                     importer.resolve(inputMBAMINDEXFileID),
                                     importer.resolve(inputFBAMFileID),
                                     importer.resolve(inputFBAMINDEXFileID),
                                     importer.resolve(inputSiblingBAMFileIDs),
                                     importer.resolve(inputSiblingBAMINDEXFileIDs),
                                     options.sample_name,
                                     options.maternal_name,
                                     options.paternal_name,
                                     options.sibling_names, options.sibling_genders, options.sibling_affected,
                                     options.bypass, options.cadd_lines,
                                     options.chrom_dir, options.edit_dir,
                                     options.split_lines, options.genome_build, options.cadd_data,
                                     cores=context.config.misc_cores,
                                     memory=context.config.misc_mem,
                                     disk=context.config.misc_disk)

            # Init the outstore
            init_job = Job.wrapJobFn(run_write_info_to_outstore, context, sys.argv,
                                     memory=context.config.misc_mem,
                                     disk=context.config.misc_disk)
            init_job.addFollowOn(root_job)            
            
            # Run the job and store the returned list of output files to download
            toil.start(init_job)
        else:
            toil.restart()
            
    end_time_pipeline = timeit.default_timer()
    run_time_pipeline = end_time_pipeline - start_time_pipeline
 
    logger.info("All jobs completed successfully. Pipeline took {} seconds.".format(run_time_pipeline))
    
